Package: IsingSampler
Type: Package
Title: Sampling Methods and Distribution Functions for the Ising Model
Version: 0.5.0
Authors@R: c(
  person("Sacha", "Epskamp", email = "mail@sachaepskamp.com",role = c("aut", "cre")),
  person("Jesse", "Boot", role = c("ctb")),
  person("Adela Maria", "Isvoranu", email = "isvoranu.adela@gmail.com", role = c("ctb"))
  )
Maintainer: Sacha Epskamp <mail@sachaepskamp.com>
Description: Sample states from the Ising model and compute the probability of
    states. Sampling can be done for any number of nodes, but due to the
    intractability of the Ising model the distribution can only be computed up
    to roughly 10 nodes. The Blume-Capel model, an Ising model with an
    additional on-site quadratic (crystal-field) term, is also supported.
License: GPL-2
Imports: Rcpp (>= 0.10.4), plyr, magrittr, nnet, dplyr
Depends: R (>= 3.0.0)
LinkingTo: Rcpp
URL: https://github.com/SachaEpskamp/IsingSampler
NeedsCompilation: yes
Packaged: 2026-07-16 03:04:33 UTC; sachaepskamp
Author: Sacha Epskamp [aut, cre],
  Jesse Boot [ctb],
  Adela Maria Isvoranu [ctb]
Repository: CRAN
Date/Publication: 2026-07-16 07:50:02 UTC

useDynLib(IsingSampler)
export(IsingSampler)
export(IsingSumLikelihood)
export(IsingLikelihood)
export(IsingStateProb)
importFrom("plyr","ddply","alply")
import(Rcpp)
export(LinTransform)
import(magrittr)
export(EstimateIsing)
export(IsingPL)
import(nnet)
export(EstimateIsing)
export(EstimateIsingPL)
export(EstimateIsingUni)
export(EstimateIsingBi)
export(EstimateIsingLL)
import(dplyr)
#export(NodeInformation)
export(IsingEntrophy)
importFrom("stats", "binomial", "coef", "glm", "loglin", "nlm")

IsingEntropy <- function(
  graph,
  thresholds,
  beta = 1,
  conditional = numeric(0), # Indices of nodes to condition on
  marginalize = numeric(0),
  base = 2,
  responses = c(0L, 1L),
  delta = 0
  ){
  stopifnot(isSymmetric(graph))
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }

  Lik <- IsingLikelihood(graph, thresholds, beta, responses, delta = delta)

  varNames <- names(Lik)[-1]

  if (any(marginalize %in% conditional)){
    stop("can not marginalize over nodes to condition on")
  }

  if (length(marginalize) > 0){
    Lik <- Lik %>%
      dplyr::group_by(dplyr::across(dplyr::all_of(varNames[-marginalize]))) %>%
      dplyr::summarise(Probability = sum(.data$Probability), .groups = "drop")
  }

  if (length(conditional) > 0){
    Lik <- Lik %>% dplyr::group_by(dplyr::across(dplyr::all_of(varNames[conditional])))
  } else {
    Lik <- Lik %>% dplyr::ungroup()
  }

  condLik <- Lik %>%
    dplyr::summarise(
      P = sum(.data$Probability),
      Entropy = -sum(.data$Probability / sum(.data$Probability) *
                        log(.data$Probability / sum(.data$Probability), base)),
      .groups = "drop"
      )

  Ent <- sum(condLik$P * condLik$Entropy)

  return(Ent)
}


# Deprecated: retained for backward compatibility (misspelled name).
IsingEntrophy <- function(...){
  .Deprecated("IsingEntropy")
  IsingEntropy(...)
}


NodeInformation <- function(
  graph,
  thresholds,
  beta = 1,
  base = 2,
  responses = c(0L, 1L),
  delta = 0
){
  stopifnot(isSymmetric(graph))
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }

  # Shannon information of node to whole graph

  sapply(seq_len(ncol(graph)),
         function(Node){
           graphEntropy <- IsingEntropy(graph, thresholds, beta, responses=responses,base=base, marginalize = Node, delta = delta)
           nodeEntropy <- IsingEntropy(graph, thresholds, beta, responses=responses,base=base, conditional = Node, delta = delta)
           return(graphEntropy - nodeEntropy)
         })
}

# Wrapper function for:
# - pl: pseudolikelihood
# - uni: univariate logistic regressions
# - bi: bivariate logistic regressions
# - ll: Loglinear model

EstimateIsing <- function(data, responses, beta = 1, method = c('uni','pl' , 'bi', 'll'),adj = matrix(1, ncol(data), ncol(data)), ...){

  method <- match.arg(method)
  
  if (!identical(adj,matrix(1, ncol(data), ncol(data)))){

    if (!method %in% c("uni","ll")) stop("Adjacency structure only supported if method = 'll' or method = 'uni'")
  }

  data <- as.matrix(data)
 
  switch(method[[1]],
        pl = EstimateIsingPL(data, responses, beta, ...)  ,
        uni = EstimateIsingUni(data, responses, beta, adj=adj,...)  ,
        bi = EstimateIsingBi(data, responses, beta, ...),
        ll = EstimateIsingLL(data, responses, beta, adj=adj,...)
         )
}

# Pseudo likelihood:
IsingPL <- function(
  x, # Vector or data frame containing data
 graph, thresholds, beta, responses = c(0L,1L))
{
  stopifnot(isSymmetric(graph))
  checkResponses(responses)
  stopifnot(length(responses)==2)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }
  N <- nrow(graph)
  
  # If x is vector, turn it into a matrix:
  if (is.vector(x)){
    x <- t(x)
  }
  
  if (is.data.frame(x)){
    x <- as.matrix(x)
  }
  
  stopifnot(is.matrix(x))
  
  # Compute pseudo likelihood:
  PL <- PseudoLikelihood(x, graph, thresholds, beta, responses, logis = TRUE)
  
  
  return(PL)
}

# Optimisation:
EstimateIsingPL <- function(data, responses, beta = 1, ...){
  if (missing(responses)){
    responses <- sort(unique(c(data)))
  }

  if (length(responses) != 2){
    stop("Binary data required")
  }

  if (!all(responses %in% c(data))){
    warning("The data do not contain both response options given in 'responses'; 'responses' should match the coding used in the data.")
  }

  # Optimisation function. The pseudo-likelihood depends on the parameters
  # only through beta * graph and beta * thresholds, so optimizing at the
  # user-supplied beta returns parameters that, combined with that beta,
  # reproduce the data-generating measure (params_beta = params_1 / beta):
  optimFun <- function(par, data){
    Ni <- ncol(data)
    stopifnot(length(par) == (Ni*(Ni-1)/2)+Ni)
    graph <- matrix(0,Ni, Ni)
    graph[upper.tri(graph)] <- par[1:(Ni*(Ni-1)/2)]
    graph[lower.tri(graph)] <- t(graph)[lower.tri(graph)]
    thresholds <- par[(Ni*(Ni-1)/2+1):length(par)]
    -2*IsingPL(data, graph, thresholds, beta, responses = responses)
  }
  
  # Run optimizer:
  Ni <- ncol(data)
#   optimRes <- optim(rep(0,(Ni*(Ni-1)/2)+Ni), optimFun, data = Data, ...)
  optimRes <- nlm(optimFun, rep(0,(Ni*(Ni-1)/2)+Ni), data = data, ...)
  
  # Cunstruct graph and thresholds:
  par <- optimRes$estimate
  graph <- matrix(0,Ni, Ni)
  graph[upper.tri(graph)] <- par[1:(Ni*(Ni-1)/2)]
  graph[lower.tri(graph)] <- t(graph)[lower.tri(graph)]
  thresholds <- par[(Ni*(Ni-1)/2+1):length(par)]
  
  return(list(
    graph = graph,
    thresholds = thresholds,
    results = optimRes))
}

# Tests for the 'constrain' argument of IsingSampler() (GitHub issue #1):
# constraints must be honored by all three sampling methods ("MH", "CFTP",
# "direct"), and for "direct" the constrained samples must follow the
# conditional distribution given the clamped nodes.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

set.seed(1234)

## Small test model -----------------------------------------------------------
N <- 4
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- runif(N * (N - 1) / 2, -0.5, 1)
graph <- graph + t(graph)
thresholds <- runif(N, -1, 1)
responses <- c(0L, 1L)

## 1) Clamping node 1 to each response option, all methods --------------------
nSamp <- 50
for (method in c("MH", "CFTP", "direct")) {
  for (val in responses) {
    constrain <- matrix(NA_real_, nSamp, N)
    constrain[, 1] <- val
    samples <- IsingSampler(nSamp, graph, thresholds, method = method,
                            responses = responses, constrain = constrain)
    samples <- as.matrix(samples)
    stopifnot(nrow(samples) == nSamp, ncol(samples) == N)
    # The constrained column must equal the clamped value in every sample:
    stopifnot(all(samples[, 1] == val))
    # Unconstrained columns must still contain valid response options:
    stopifnot(all(as.vector(samples[, -1]) %in% responses))
  }
}

## 2) Mixed constraints: different rows constrained differently (direct) ------
constrain <- matrix(NA_real_, 60, N)
constrain[1:20, 1] <- 0   # rows 1-20: node 1 clamped to 0
constrain[21:40, 1] <- 1  # rows 21-40: node 1 clamped to 1
constrain[21:40, 3] <- 0  # rows 21-40: node 3 also clamped to 0
                          # rows 41-60: unconstrained
samples <- as.matrix(IsingSampler(60, graph, thresholds, method = "direct",
                                  constrain = constrain))
stopifnot(all(samples[1:20, 1] == 0))
stopifnot(all(samples[21:40, 1] == 1))
stopifnot(all(samples[21:40, 3] == 0))
stopifnot(all(as.vector(samples) %in% responses))

## 3) Conditional distribution is respected for method = "direct" -------------
# 3-node model: compare empirical state frequencies of constrained direct
# samples against the exact conditional probabilities obtained by restricting
# IsingLikelihood() output to the constrained subset.
set.seed(5678)
N3 <- 3
graph3 <- matrix(0, N3, N3)
graph3[upper.tri(graph3)] <- c(0.5, -0.3, 0.8)
graph3 <- graph3 + t(graph3)
thresholds3 <- c(-0.2, 0.1, 0.3)

lik <- IsingLikelihood(graph3, thresholds3, beta = 1, responses = c(0L, 1L))
# Restrict to states with node 1 == 1 and renormalize:
sub <- lik[lik[, 2] == 1, ]
condP <- sub$Probability / sum(sub$Probability)
condKey <- apply(sub[, -1], 1, paste, collapse = ",")

nBig <- 20000
constrain3 <- matrix(NA_real_, nBig, N3)
constrain3[, 1] <- 1
samples3 <- as.matrix(IsingSampler(nBig, graph3, thresholds3, method = "direct",
                                   constrain = constrain3))
stopifnot(all(samples3[, 1] == 1))
empKey <- apply(samples3, 1, paste, collapse = ",")
empFreq <- table(factor(empKey, levels = condKey)) / nBig
# All conditional states should be covered and frequencies close to theory:
stopifnot(max(abs(as.numeric(empFreq) - condP)) < 0.02)

## 4) Validation of the 'constrain' argument ----------------------------------
# Wrong dimensions must give an error, for every method:
for (method in c("MH", "CFTP", "direct")) {
  res <- try(IsingSampler(10, graph, thresholds, method = method,
                          constrain = matrix(NA_real_, 5, N)), silent = TRUE)
  stopifnot(inherits(res, "try-error"))
}
# Values not in 'responses' must give an error:
badConstrain <- matrix(NA_real_, 10, N)
badConstrain[, 1] <- 2
res <- try(IsingSampler(10, graph, thresholds, constrain = badConstrain),
           silent = TRUE)
stopifnot(inherits(res, "try-error"))

cat("All constrain tests passed.\n")

IsingSumLikelihood <- function(graph, thresholds, beta, responses = c(0L,1L), delta = 0)
{
  if (!is.matrix(graph) || !is.numeric(graph) || nrow(graph) != ncol(graph))
  {
    stop("'graph' must be a numeric square matrix.")
  }
  stopifnot(isSymmetric(graph))
  if (length(thresholds) != 1L && length(thresholds) != ncol(graph))
  {
    stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  }
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }
  N <- nrow(graph)
  if (length(thresholds) == 1L) thresholds <- rep(thresholds, N)
  if (length(delta) == 1L) delta <- rep(delta, N)
  Allstates <- do.call(expand.grid,lapply(1:N,function(x)responses))
  P <- exp(- beta * apply(Allstates,1,function(s)H(graph,s,thresholds,delta)))
  SumScores <- rowSums(Allstates)

  df <- plyr::ddply(data.frame(Sum = SumScores, P = P),"Sum",plyr::summarize,P=sum(P))
  df$P <- df$P / sum(df$P)
  return(df)
}

IsingLikelihood <- function(graph, thresholds, beta, responses = c(0L,1L), potential = FALSE, delta = 0)
{
  if (!is.matrix(graph) || !is.numeric(graph) || nrow(graph) != ncol(graph))
  {
    stop("'graph' must be a numeric square matrix.")
  }
  stopifnot(isSymmetric(graph))
  if (length(thresholds) != 1L && length(thresholds) != ncol(graph))
  {
    stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  }
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }
  N <- nrow(graph)
  if (length(thresholds) == 1L) thresholds <- rep(thresholds, N)
  if (length(delta) == 1L) delta <- rep(delta, N)
  Allstates <- do.call(expand.grid,lapply(1:N,function(x)responses))
  P <- exp(- beta * apply(Allstates,1,function(s)H(graph,s,thresholds,delta)))
  if (potential){
    df <- cbind(Potential = P, Allstates)    
  } else {
    df <- cbind(Probability = P / sum(P), Allstates)
  }

  return(df)
}

IsingStateProb <- function(s,graph,thresholds,beta,responses=c(0L,1L), delta = 0)
{
  if (!is.matrix(graph) || !is.numeric(graph) || nrow(graph) != ncol(graph))
  {
    stop("'graph' must be a numeric square matrix.")
  }
  stopifnot(isSymmetric(graph))
  checkResponses(responses)
  if (!is.list(s)) s <- list(s)
  N <- length(s[[1]])
  if (N != ncol(graph))
  {
    stop("Each state in 's' must have length equal to the number of nodes (ncol(graph)).")
  }
  if (length(thresholds) != 1L && length(thresholds) != ncol(graph))
  {
    stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  }
  if (length(thresholds) == 1L) thresholds <- rep(thresholds, N)
  if (length(delta) == 1L) delta <- rep(delta, N)
  Allstates <- do.call(expand.grid,lapply(1:N,function(x)responses))
  Dist <- exp(- beta * apply(Allstates,1,function(s)H(graph,s,thresholds,delta)))
  Z <- sum(Dist)

  sapply(s, function(x)exp(-beta*H(graph,x,thresholds,delta))/Z)
}

# Tests for the non-regularized estimators (EstimateIsing*):
# (a) the 'beta' argument must actually be used. Convention: the Ising
#     measure depends on the parameters only through beta * graph and
#     beta * thresholds, so all estimators return the beta = 1 estimates
#     divided by beta (for 'pl' this is achieved by optimizing the
#     pseudo-likelihood at the supplied beta, which yields the same scaling);
# (b) supplying 'responses' explicitly must give the same result as the
#     default (missing) call on matching data;
# (c) EstimateIsingPL() must warn when a supplied 'responses' coding does
#     not match the values present in the data.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

## Small seeded 3-node test model ---------------------------------------------
set.seed(123)
N <- 3
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- c(0.4, -0.2, 0.6)
graph <- graph + t(graph)
thresholds <- c(-0.1, 0.1, 0)
data <- as.matrix(IsingSampler(500, graph, thresholds, responses = c(0L, 1L)))

## (a) beta rescaling ----------------------------------------------------------

# Pseudo-likelihood: fitting at beta = 2 optimizes the same objective as a
# function of 2 * par, so the estimates must (up to nlm convergence noise)
# equal the beta = 1 estimates divided by 2. Generous tolerance for nlm:
pl1 <- EstimateIsingPL(data, beta = 1)
pl2 <- EstimateIsingPL(data, beta = 2)
stopifnot(max(abs(pl2$graph - pl1$graph / 2)) < 1e-3)
stopifnot(max(abs(pl2$thresholds - pl1$thresholds / 2)) < 1e-3)

# Univariate logistic regressions: rescaling is exact (deterministic glm):
uni1 <- EstimateIsingUni(data, beta = 1)
uni2 <- EstimateIsingUni(data, beta = 2)
stopifnot(isTRUE(all.equal(uni2$graph, uni1$graph / 2)))
stopifnot(isTRUE(all.equal(uni2$thresholds, uni1$thresholds / 2)))

# Loglinear model: rescaling is exact (deterministic loglin/IPF):
ll1 <- EstimateIsingLL(data, beta = 1)
ll2 <- EstimateIsingLL(data, beta = 2)
stopifnot(isTRUE(all.equal(ll2$graph, ll1$graph / 2)))
stopifnot(isTRUE(all.equal(ll2$thresholds, ll1$thresholds / 2)))

# Bivariate (multinomial) regressions can be numerically unstable, so only
# the beta rescaling arithmetic is exercised here: identical RNG seeds make
# the two multinom fits start identically, so the estimates must satisfy the
# same divide-by-beta relation (generous tolerance for the optimizer):
set.seed(1)
bi1 <- EstimateIsingBi(data, beta = 1, trace = FALSE)
set.seed(1)
bi2 <- EstimateIsingBi(data, beta = 2, trace = FALSE)
stopifnot(all(is.finite(bi2$graph)), all(is.finite(bi2$thresholds)))
stopifnot(max(abs(bi2$graph - bi1$graph / 2)) < 1e-3)
stopifnot(max(abs(bi2$thresholds - bi1$thresholds / 2)) < 1e-3)

# Cross-method sanity check of the convention: beta * estimates should be on
# the same (measure) scale regardless of beta, e.g. 2 * pl2 approx 1 * pl1:
stopifnot(max(abs(2 * pl2$graph - pl1$graph)) < 2e-3)

## (b) explicit responses equal the default on matching 0/1 data ---------------
plDef <- EstimateIsingPL(data)
plExp <- EstimateIsingPL(data, responses = c(0, 1))
stopifnot(identical(plDef$graph, plExp$graph))
stopifnot(identical(plDef$thresholds, plExp$thresholds))

uniDef <- EstimateIsingUni(data)
uniExp <- EstimateIsingUni(data, responses = c(0, 1))
stopifnot(identical(uniDef$graph, uniExp$graph))
stopifnot(identical(uniDef$thresholds, uniExp$thresholds))

## (c) warning when 'responses' does not match the data ------------------------
w <- tryCatch(EstimateIsingPL(data, responses = c(0, 2)),
              warning = function(w) w)
stopifnot(inherits(w, "warning"))
stopifnot(grepl("response options", conditionMessage(w)))

cat("All estimator tests passed.\n")

# Tests for the return type of IsingSampler().
# All three sampling methods ("MH", "CFTP", "direct") must return an unnamed
# matrix with n rows and ncol(graph) columns. Integer-valued response options
# yield integer storage; non-integer response options yield double storage.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

set.seed(4321)

## Small test model -----------------------------------------------------------
N <- 4
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- runif(N * (N - 1) / 2, -0.5, 1)
graph <- graph + t(graph)
thresholds <- runif(N, -1, 1)

nSamp <- 30

## 1) Integer responses: matrix, no dimnames, integer storage, right dims -----
for (method in c("MH", "CFTP", "direct")) {
  samples <- IsingSampler(nSamp, graph, thresholds, method = method,
                          responses = c(0L, 1L))
  stopifnot(is.matrix(samples))
  stopifnot(is.null(dimnames(samples)))
  stopifnot(storage.mode(samples) == "integer")
  stopifnot(nrow(samples) == nSamp, ncol(samples) == N)
}

## 2) Non-integer responses: matrix, no dimnames, double storage --------------
# CFTP is binary-only; non-integer c(0, 0.5) is still two response options so
# it is valid for all three methods.
for (method in c("MH", "CFTP", "direct")) {
  samples <- IsingSampler(nSamp, graph, thresholds, method = method,
                          responses = c(0, 0.5))
  stopifnot(is.matrix(samples))
  stopifnot(is.null(dimnames(samples)))
  stopifnot(storage.mode(samples) == "double")
  stopifnot(nrow(samples) == nSamp, ncol(samples) == N)
}

## 3) direct method WITH a constrain matrix still returns a proper matrix ------
constrain <- matrix(NA_real_, nSamp, N)
constrain[, 1] <- 1L
samples <- IsingSampler(nSamp, graph, thresholds, method = "direct",
                        responses = c(0L, 1L), constrain = constrain)
stopifnot(is.matrix(samples))
stopifnot(is.null(dimnames(samples)))
stopifnot(storage.mode(samples) == "integer")
stopifnot(nrow(samples) == nSamp, ncol(samples) == N)
stopifnot(all(samples[, 1] == 1L))

cat("All return-type tests passed.\n")

# Internal helper: validate the 'responses' argument.
# Requires at least two numeric response options. Real-valued responses such as
# seq(-1, 1, by = 0.5) are allowed; states are treated as their numeric values.
checkResponses <- function(responses)
{
  if (length(responses) < 2)
  {
    stop("'responses' must contain at least two response options.")
  }
  if (anyNA(responses) || !is.numeric(responses))
  {
    stop("'responses' must be numeric.")
  }
  invisible(responses)
}

# Wrapper function:
IsingSampler <- function(n, graph, thresholds, beta = 1, nIter = 100, responses = c(0L, 1L), method = c("MH","CFTP","direct"), CFTPretry = 10, constrain, delta = 0)
{
  if (missing(graph)) stop("'graph' must be supplied.")
  if (missing(thresholds)) stop("'thresholds' must be supplied.")
  if (!is.matrix(graph) || !is.numeric(graph) || nrow(graph) != ncol(graph))
  {
    stop("'graph' must be a numeric square matrix.")
  }
  if (!isSymmetric(graph)) stop("'graph' must be a symmetric matrix.")
  if (length(n) != 1L || !is.numeric(n) || !is.finite(n) || n != round(n) || n < 1)
  {
    stop("'n' must be a single positive integer.")
  }
  if (length(nIter) != 1L || !is.numeric(nIter) || !is.finite(nIter) || nIter != round(nIter) || nIter < 1)
  {
    stop("'nIter' must be a single positive integer.")
  }
  if (length(beta) != 1L || !is.numeric(beta) || !is.finite(beta))
  {
    stop("'beta' must be a single finite numeric value.")
  }
  if (length(thresholds) != 1L && length(thresholds) != ncol(graph))
  {
    stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  }
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }
  N <- ncol(graph)

  # Recycle single-value thresholds / delta to one value per node:
  if (length(thresholds) == 1L) thresholds <- rep(thresholds, N)
  if (length(delta) == 1L) delta <- rep(delta, N)
  if (length(thresholds) != N) stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  if (length(delta) != N) stop("'delta' must have length 1 or the number of nodes (ncol(graph)).")

  method <- method[1]
  if (! method %in% c("MH","CFTP","direct")) stop("method must be 'MH', 'CFTP', or 'direct'")

  # The Blume-Capel quadratic (delta) term only has an identifiable effect with
  # more than two response options: for two options delta * s^2 is an affine
  # function of s and is absorbed into the thresholds (a constant for symmetric
  # responses such as c(-1, 1)).
  if (any(delta != 0) && length(responses) == 2)
  {
    warning("With two response options the 'delta' term is not separately identifiable from the thresholds; results are equivalent to an Ising model with adjusted thresholds.")
  }

  # Exact sampling (CFTP) is only implemented for binary (two-level) responses:
  if (method == "CFTP" && length(responses) != 2)
  {
    stop("method = 'CFTP' (exact sampling) is currently only implemented for two response options. Use method = 'MH' or method = 'direct' for more than two response options.")
  }

  # Check constrains:
  if (missing(constrain))
  {
    constrain <- matrix(NA_real_,n,ncol(graph))
  }
  if (!is.matrix(constrain) || nrow(constrain) != n || ncol(constrain) != N)
  {
    stop(sprintf("'constrain' must be an n x N matrix (%d x %d: number of samples by number of nodes), with NA indicating unconstrained values.", n, N))
  }
  if (!all(constrain[!is.na(constrain)] %in% responses))
  {
    stop("All non-NA values in 'constrain' must be members of 'responses'.")
  }
  # responses, constrain and delta are passed to C++ as doubles so that
  # non-integer response options (e.g. seq(-1, 1, by = 0.5)) are preserved:
  storage.mode(constrain) <- "double"

  if (method %in% c("MH","CFTP"))
  {
    try <- 1

    repeat{
      Res <- IsingSamplerCpp(as.integer(n), graph, thresholds, beta, as.integer(nIter), as.numeric(responses), as.logical(method == "CFTP"), constrain, as.numeric(delta))

      if (any(is.na(Res)) & method == "CFTP")
      {
        if (try < CFTPretry)
        {
          message("Restarting CFTP chain, attempt: ", try)
          try <- try + 1
        } else
        {
          warning(paste("NA's detected, this means that no CFTP sample was drawn after 100 couplings to the past and",CFTPretry,"resets of the chain. Use higher nIter value or method='MH' for inexact sampling."))
          break
        }
      } else break
    }

    # Preserve integer storage when all responses are integer-valued (keeps the
    # binary / integer behaviour identical to previous versions):
    if (all(responses == round(responses))) storage.mode(Res) <- "integer"
  } else
  {
    Res <- IsingDir(n, graph, thresholds, beta, responses, delta, constrain)
  }

  return(Res)
}

## direct sampling function:
# 'constrain' is an n x N matrix; NA = free, non-NA = clamp that node to that
# value in that sample. Constrained samples are drawn from the conditional
# distribution given the clamped nodes: the full state space is restricted to
# the states matching all non-NA entries of the constraint row and the
# probabilities are renormalized over that subset.
IsingDir <- function(n, graph, thresholds, beta,responses = c(0L,1L), delta = 0, constrain)
{
  if (!is.matrix(graph) || !is.numeric(graph) || nrow(graph) != ncol(graph))
  {
    stop("'graph' must be a numeric square matrix.")
  }
  if (!isSymmetric(graph)) stop("'graph' must be a symmetric matrix.")
  if (length(thresholds) != 1L && length(thresholds) != ncol(graph))
  {
    stop("'thresholds' must have length 1 or the number of nodes (ncol(graph)).")
  }
  checkResponses(responses)
  if (any(diag(graph)!=0))
  {
    diag(graph) <- 0
    warning("Diagonal set to 0")
  }
  N <- nrow(graph)
  if (length(thresholds) == 1L) thresholds <- rep(thresholds, N)
  if (length(delta) == 1L) delta <- rep(delta, N)
  Allstates <- do.call(expand.grid,lapply(1:N,function(x)responses))
  P <- exp(- beta * apply(Allstates,1,function(s)H(graph,s,thresholds,delta)))

  # Unconstrained sampling (also when all entries of 'constrain' are NA):
  if (missing(constrain) || all(is.na(constrain)))
  {
    Res <- Allstates[sample(1:nrow(Allstates),n,TRUE,prob=P), , drop = FALSE]
  } else
  {
    if (!is.matrix(constrain) || nrow(constrain) != n || ncol(constrain) != N)
    {
      stop(sprintf("'constrain' must be an n x N matrix (%d x %d: number of samples by number of nodes), with NA indicating unconstrained values.", n, N))
    }
    if (!all(constrain[!is.na(constrain)] %in% responses))
    {
      stop("All non-NA values in 'constrain' must be members of 'responses'.")
    }

    # Group identical constraint rows so the restricted distribution is computed
    # only once per unique constraint pattern:
    key <- apply(constrain, 1, paste, collapse = "\r")
    stateIndex <- integer(n)
    for (k in unique(key))
    {
      rows <- which(key == k)
      cons <- constrain[rows[1], ]
      fixed <- which(!is.na(cons))
      if (length(fixed) == 0)
      {
        sub <- seq_len(nrow(Allstates))
      } else
      {
        keep <- rep(TRUE, nrow(Allstates))
        for (j in fixed) keep <- keep & (Allstates[[j]] == cons[j])
        sub <- which(keep)
      }
      # Renormalization to the restricted subset happens implicitly in sample():
      stateIndex[rows] <- sub[sample.int(length(sub), length(rows), replace = TRUE, prob = P[sub])]
    }
    Res <- Allstates[stateIndex, , drop = FALSE]
  }

  # Return an unnamed matrix, matching the MH/CFTP methods (which return a
  # matrix from C++). Integer storage is preserved when all responses are
  # integer-valued, mirroring the MH/CFTP branch of IsingSampler():
  Res <- as.matrix(Res)
  dimnames(Res) <- NULL
  if (all(responses == round(responses))) storage.mode(Res) <- "integer"
  return(Res)
}

# Wrapper around IsingSampler for the Blume-Capel model.
#
# The Blume-Capel model is an Ising model with an additional on-site quadratic
# ("single-ion anisotropy" / crystal-field) term in the Hamiltonian:
#
#   H = - sum_i tau_i s_i - sum_{i<j} omega_ij s_i s_j + sum_i delta_i s_i^2
#
# Compared with IsingSampler() the only new ingredient is 'delta'. It is a true
# extension only with more than two (ordered) response options; the default
# response options are the spin-1 values c(-1, 0, 1). With delta_i > 0 the middle
# category is favoured; with delta_i < 0 the extreme categories are favoured.
#
# This is a thin wrapper: it sets the default responses and a required 'delta'
# and forwards everything to IsingSampler(). Exact sampling ("CFTP") is binary
# only and therefore not offered here (delta has no identifiable effect on binary
# responses anyway); use method = "MH" (default) or "direct".
BlumeCapelSampler <- function(n, graph, thresholds, delta, beta = 1, nIter = 100,
                              responses = c(-1L, 0L, 1L),
                              method = c("MH", "direct"), CFTPretry = 10, constrain)
{
  if (missing(delta))
  {
    stop("'delta' must be supplied for the Blume-Capel model (use IsingSampler() for the delta = 0 / Ising case).")
  }
  method <- match.arg(method)

  args <- list(n = n, graph = graph, thresholds = thresholds, beta = beta,
               nIter = nIter, responses = responses, method = method,
               CFTPretry = CFTPretry, delta = delta)
  if (!missing(constrain)) args$constrain <- constrain

  do.call(IsingSampler, args)
}

# Tests for input validation of the exported model functions.
# IsingSampler() previously used an OR guard so supplying only 'graph' or only
# 'thresholds' passed and then failed with an obscure error. These tests assert
# the specific, clear error messages and confirm valid calls still work.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

set.seed(1)

## Small valid model ----------------------------------------------------------
N <- 3
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- runif(N * (N - 1) / 2, -0.5, 1)
graph <- graph + t(graph)
thresholds <- runif(N, -1, 1)

# Helper: run 'expr' and return the error message (or NA if no error).
errMsg <- function(expr) tryCatch({ force(expr); NA_character_ },
                                  error = function(e) conditionMessage(e))

## 1) Missing graph -----------------------------------------------------------
stopifnot(errMsg(IsingSampler(10, thresholds = thresholds)) == "'graph' must be supplied.")

## 2) Missing thresholds ------------------------------------------------------
stopifnot(errMsg(IsingSampler(10, graph = graph)) == "'thresholds' must be supplied.")

## 3) Non-square graph --------------------------------------------------------
badGraph <- matrix(0, 3, 2)
stopifnot(errMsg(IsingSampler(10, badGraph, c(0, 0))) == "'graph' must be a numeric square matrix.")

## 4) Threshold length mismatch -----------------------------------------------
stopifnot(errMsg(IsingSampler(10, graph, c(0, 0))) ==
            "'thresholds' must have length 1 or the number of nodes (ncol(graph)).")

## 5) Non-integer / negative n ------------------------------------------------
stopifnot(errMsg(IsingSampler(2.5, graph, thresholds)) == "'n' must be a single positive integer.")
stopifnot(errMsg(IsingSampler(-1, graph, thresholds)) == "'n' must be a single positive integer.")
stopifnot(errMsg(IsingSampler(c(1, 2), graph, thresholds)) == "'n' must be a single positive integer.")

## 6) Invalid beta ------------------------------------------------------------
stopifnot(errMsg(IsingSampler(10, graph, thresholds, beta = c(1, 2))) ==
            "'beta' must be a single finite numeric value.")
stopifnot(errMsg(IsingSampler(10, graph, thresholds, beta = Inf)) ==
            "'beta' must be a single finite numeric value.")

## 7) Invalid nIter -----------------------------------------------------------
stopifnot(errMsg(IsingSampler(10, graph, thresholds, nIter = 2.5)) ==
            "'nIter' must be a single positive integer.")

## 8) IsingStateProb with asymmetric graph ------------------------------------
asym <- graph
asym[1, 2] <- asym[1, 2] + 1  # break symmetry
msg <- errMsg(IsingStateProb(rep(0L, N), asym, thresholds, beta = 1))
stopifnot(grepl("symmetric", msg) || grepl("isSymmetric", msg))

## 9) IsingStateProb with wrong state length ----------------------------------
stopifnot(errMsg(IsingStateProb(rep(0L, N + 1), graph, thresholds, beta = 1)) ==
            "Each state in 's' must have length equal to the number of nodes (ncol(graph)).")

## 10) Distribution functions reject non-square graph -------------------------
stopifnot(errMsg(IsingLikelihood(badGraph, c(0, 0), beta = 1)) ==
            "'graph' must be a numeric square matrix.")
stopifnot(errMsg(IsingSumLikelihood(badGraph, c(0, 0), beta = 1)) ==
            "'graph' must be a numeric square matrix.")

## Valid calls still work (smoke tests) ---------------------------------------
samples <- IsingSampler(10, graph, thresholds)
stopifnot(is.matrix(samples), nrow(samples) == 10, ncol(samples) == N)

samplesDir <- IsingSampler(10, graph, thresholds, method = "direct")
stopifnot(is.matrix(samplesDir), nrow(samplesDir) == 10, ncol(samplesDir) == N)

lik <- IsingLikelihood(graph, thresholds, beta = 1)
stopifnot(is.data.frame(lik))

sumLik <- IsingSumLikelihood(graph, thresholds, beta = 1)
stopifnot(is.data.frame(sumLik))

p <- IsingStateProb(rep(0L, N), graph, thresholds, beta = 1)
stopifnot(is.numeric(p), length(p) == 1)

## 11) LinTransform argument validation ---------------------------------------
# Wrong 'from' / 'to' lengths must error clearly.
stopifnot(errMsg(LinTransform(graph, thresholds, from = c(0L, 1L, 2L))) ==
            "'from' and 'to' must each be vectors of length 2.")
stopifnot(errMsg(LinTransform(graph, thresholds, to = 1L)) ==
            "'from' and 'to' must each be vectors of length 2.")

# Supplying exactly one of 'a' / 'b' must error.
stopifnot(errMsg(LinTransform(graph, thresholds, a = 1)) ==
            "Supply both 'a' and 'b', or neither.")
stopifnot(errMsg(LinTransform(graph, thresholds, b = 0)) ==
            "Supply both 'a' and 'b', or neither.")

# A valid call still works.
lt <- LinTransform(graph, thresholds)
stopifnot(is.list(lt), all(c("graph", "thresholds") %in% names(lt)))

cat("All validation tests passed.\n")

# Tests for IsingPL() (pseudo-likelihood):
# (a) 'beta' must actually be used (previously ignored in the C++ code) and
#     the result must equal the manually computed pseudo-log-likelihood;
# (b) real-valued response options must be preserved (previously silently
#     truncated to integers by an IntegerVector parameter);
# (c) the default binary 0/1, beta = 1 path must reproduce the pre-fix value.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

# Manual R implementation of the node-conditional pseudo-log-likelihood:
# logPL = sum_i sum_j [ beta * x_ij * e_ij
#                       - log( exp(beta*r1*e_ij) + exp(beta*r2*e_ij) ) ]
# with e_ij = thresholds_j + sum_jj x_i,jj * graph_j,jj
plR <- function(x, graph, thresholds, beta, responses) {
  logPL <- 0
  for (i in seq_len(nrow(x))) {
    for (j in seq_len(ncol(x))) {
      e <- thresholds[j] + sum(x[i, ] * graph[j, ])
      logPL <- logPL + beta * x[i, j] * e -
        log(exp(beta * responses[1] * e) + exp(beta * responses[2] * e))
    }
  }
  logPL
}

## Small 3-node test model -----------------------------------------------------
set.seed(42)
N <- 3
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- c(0.5, -0.3, 0.8)
graph <- graph + t(graph)
thresholds <- c(-0.2, 0.1, 0.3)
x <- as.matrix(IsingSampler(10, graph, thresholds, responses = c(0L, 1L)))

## (a) beta is used and matches the manual R computation -----------------------
pl1 <- IsingPL(x, graph, thresholds, beta = 1)
pl2 <- IsingPL(x, graph, thresholds, beta = 2)
stopifnot(pl1 != pl2)
stopifnot(isTRUE(all.equal(pl1, plR(x, graph, thresholds, 1, c(0, 1)))))
stopifnot(isTRUE(all.equal(pl2, plR(x, graph, thresholds, 2, c(0, 1)))))

## (b) non-integer response options are preserved ------------------------------
xHalf <- x * 0.5  # data on the c(0, 0.5) response scale
plHalf  <- IsingPL(xHalf, graph, thresholds, beta = 1, responses = c(0, 0.5))
plSeven <- IsingPL(xHalf, graph, thresholds, beta = 1, responses = c(0, 0.7))
stopifnot(plHalf != plSeven)
stopifnot(isTRUE(all.equal(plHalf,
                           plR(xHalf, graph, thresholds, 1, c(0, 0.5)))))
stopifnot(isTRUE(all.equal(plSeven,
                           plR(xHalf, graph, thresholds, 1, c(0, 0.7)))))

## (c) backward compatibility: binary 0/1 with beta = 1 ------------------------
# Pre-fix reference value computed with IsingSampler before this fix on the
# exact same seeded example (seed 42, model and data as above):
preFixReference <- -16.420410068413286
stopifnot(isTRUE(all.equal(pl1, preFixReference)))

## Numerical stability: large beta must not overflow to NaN/Inf ----------------
plBig <- IsingPL(x, graph, thresholds, beta = 500)
stopifnot(is.finite(plBig))

cat("All pseudo-likelihood tests passed.\n")

# Tests for IsingEntropy() and its deprecated alias IsingEntrophy().
# On a small 3-node model, IsingEntropy() must return a finite, non-negative
# number, and the deprecated IsingEntrophy() must return the identical value
# while emitting a deprecation warning.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

set.seed(4321)

## Small test model -----------------------------------------------------------
N <- 3
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- runif(N * (N - 1) / 2, -0.5, 1)
graph <- graph + t(graph)
thresholds <- runif(N, -1, 1)

## 1) IsingEntropy() returns a finite non-negative number ----------------------
ent <- IsingEntropy(graph, thresholds)
stopifnot(is.numeric(ent), length(ent) == 1L)
stopifnot(is.finite(ent))
stopifnot(ent >= 0)

## 2) The deprecated IsingEntrophy() returns the identical value ---------------
entOld <- suppressWarnings(IsingEntrophy(graph, thresholds))
stopifnot(isTRUE(all.equal(ent, entOld)))

## 3) IsingEntrophy() emits a (deprecation) warning ---------------------------
warned <- FALSE
invisible(withCallingHandlers(
  IsingEntrophy(graph, thresholds),
  warning = function(w) {
    warned <<- TRUE
    invokeRestart("muffleWarning")
  }
))
stopifnot(warned)

## 4) delta != 0 exercises the Blume-Capel path and changes the entropy --------
## Use three (non-symmetric) response options so 'delta' has an identifiable
## effect (with two response options delta is absorbed into the thresholds).
graph3 <- matrix(0, N, N)
graph3[upper.tri(graph3)] <- runif(N * (N - 1) / 2, -0.5, 1)
graph3 <- graph3 + t(graph3)
thr3 <- runif(N, -1, 1)

ent0   <- IsingEntropy(graph3, thr3, responses = c(0L, 1L, 2L), delta = 0)
entHalf <- IsingEntropy(graph3, thr3, responses = c(0L, 1L, 2L), delta = 0.5)
stopifnot(is.finite(ent0), is.finite(entHalf))
stopifnot(!isTRUE(all.equal(ent0, entHalf)))

## 5) NodeInformation() returns non-negative mutual information -----------------
NI <- NodeInformation(graph, thresholds)
stopifnot(is.numeric(NI), length(NI) == ncol(graph))
stopifnot(all(is.finite(NI)))
# Mutual information is non-negative (allow tiny numerical tolerance):
stopifnot(all(NI > -1e-10))

cat("All entropy tests passed.\n")

# Distributional tests: sampler state distributions vs exact probabilities,
# and consistency of the distribution functions (IsingStateProb,
# IsingSumLikelihood, LinTransform).
library("IsingSampler")

set.seed(1234)

## Fixed 4-node model (0/1 coding) ------------------------------------------
N <- 4
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- c(0.5, -0.4, 0.3, 0.2, -0.1, 0.6)
graph <- graph + t(graph)
thresholds <- c(-0.2, 0.1, -0.5, 0.3)
beta <- 1
responses <- c(0L, 1L)

## Exact distribution over all 2^4 states:
lik <- IsingLikelihood(graph, thresholds, beta, responses)
stopifnot(nrow(lik) == 2^N)
stopifnot(abs(sum(lik$Probability) - 1) < 1e-10)

# Key identifying each state (columns 2:(N+1) of the IsingLikelihood output
# hold the state configurations):
stateKey <- function(states) apply(states, 1, paste, collapse = "")
exactKey <- stateKey(lik[, -1])
exactP <- lik$Probability

## 1. Agreement of the three samplers with the exact distribution -----------
# n = 5000 samples per method; empirical state frequencies must be within
# 0.05 (max absolute deviation) of the exact probabilities. Note that
# method = "MH" is approximate (finite chain, nIter = 200), hence the
# generous tolerance; "direct" and "CFTP" are exact samplers and only carry
# Monte Carlo error.
n <- 5000
maxDev <- function(sample) {
  freq <- table(factor(stateKey(sample), levels = exactKey)) / nrow(sample)
  max(abs(as.numeric(freq) - exactP))
}

devs <- c(
  direct = maxDev(IsingSampler(n, graph, thresholds, beta, responses = responses,
                               method = "direct")),
  CFTP = maxDev(IsingSampler(n, graph, thresholds, beta, responses = responses,
                             method = "CFTP")),
  MH = maxDev(IsingSampler(n, graph, thresholds, beta, nIter = 200,
                           responses = responses, method = "MH"))
)
print(round(devs, 4))
stopifnot(all(devs < 0.05))

## 2. IsingStateProb sums to 1 over the full state space --------------------
allStates <- as.matrix(do.call(expand.grid, replicate(N, responses,
                                                      simplify = FALSE)))
pAll <- IsingStateProb(lapply(seq_len(nrow(allStates)),
                              function(i) allStates[i, ]),
                       graph, thresholds, beta, responses)
stopifnot(length(pAll) == 2^N)
stopifnot(abs(sum(pAll) - 1) < 1e-10)

## 3. LinTransform round-trip invariance ------------------------------------
# Transform the 0/1 parameters to -1/1 coding; state s (0/1) corresponds to
# state 2*s - 1 (-1/1). Both codings must assign identical probabilities to
# corresponding states.
tr <- LinTransform(graph, thresholds, from = c(0L, 1L), to = c(-1L, 1L))
lik2 <- IsingLikelihood(tr$graph, tr$thresholds, beta, responses = c(-1L, 1L))

# expand.grid enumerates both codings in the same order and 0 -> -1, 1 -> 1
# preserves that order; verify the state mapping explicitly:
stopifnot(all(as.matrix(lik2[, -1]) == 2 * as.matrix(lik[, -1]) - 1))
stopifnot(isTRUE(all.equal(lik$Probability, lik2$Probability,
                           tolerance = 1e-12)))

## 4. IsingSumLikelihood probabilities sum to 1 ------------------------------
# IsingSumLikelihood returns a data frame with columns 'Sum' (sum score) and
# 'P' (probability of that sum score); the probabilities must sum to 1.
sl <- IsingSumLikelihood(graph, thresholds, beta, responses)
stopifnot(is.data.frame(sl), all(c("Sum", "P") %in% names(sl)))
stopifnot(nrow(sl) == N + 1) # sum scores 0:4 for 0/1 responses
stopifnot(all(sl$P >= 0))
stopifnot(abs(sum(sl$P) - 1) < 1e-10)

# The sum-score distribution must also match the aggregated exact state
# distribution:
sumExact <- tapply(exactP, rowSums(lik[, -1]), sum)
stopifnot(isTRUE(all.equal(unname(sl$P), unname(as.numeric(sumExact)),
                           tolerance = 1e-10)))

cat("test-distribution.R: all tests passed\n")

# Univariate:
EstimateIsingUni <- function(data, responses, beta = 1, adj = matrix(1, ncol(data), ncol(data)), min_sum = -Inf, thresholding = FALSE, alpha = 0.01, AND = TRUE, ...){
  data <- as.matrix(data)

  # Check data. 'min_sum' refers to sum scores on the original response
  # coding, so this check must run on the raw data (before binarization):
  if (min_sum > -Inf){
    if (min(rowSums(data)) < min_sum){
      stop("One or more sumscores in the data are lower than the threshold set using the 'min_sum' argument.")
    }
  }

  if (missing(responses)){
    responses <- sort(unique(c(data)))
  }

  if (length(responses) != 2){
    stop("Binary data required")
  }

  if (!is.logical(adj)){
    adj <- adj != 0
  }
  diag(adj) <- FALSE

  # Keep the raw data: the minimum sum-score correction below operates on
  # the original response coding, not on the binarized 0/1 recoding:
  raw <- data
  r <- sort(responses)

  # Rescale data to binary:
  binarize <- function(x, responses){
    x2 <- x
    x2[x==responses[1]] <- 0
    x2[x==responses[2]] <- 1
    x2
  }
  data <- binarize(data, responses)

  # Number of variables:
  n <- ncol(data)

  # GLM for every node:
  Res <- lapply(seq_len(n), function(i){
    # Minimum sum-score correction, on the RAW response coding: a row is
    # biased for node i iff the rest score is compatible with the sum-score
    # truncation when node i takes the high response but not when it takes
    # the low response (for 0/1 coding this reduces to rest == min_sum - 1).
    # With the default min_sum = -Inf this condition is all-FALSE, so no
    # rows are dropped:
    rest_raw <- rowSums(raw[,-i, drop = FALSE])
    biased <- rest_raw + r[2] >= min_sum & rest_raw + r[1] < min_sum
    data <- data[!biased, , drop = FALSE]
    glm(data[,i] ~ data[,adj[i,]], family = binomial, ...)
  })
  
  # Coefficients:
  Coefs <- lapply(Res, coef)
  
  # Thresholds:
  Thresholds <- sapply(Coefs, '[[', 1)
  
  # P-values:
  p_values <- lapply(Res, function(model) {
    coef_summary <- summary(model)$coefficients
    return(coef_summary[, "Pr(>|z|)"])
  })
  # Raw network:
  Raw_Net <- matrix(0, n, n)
  for (i in seq_len(n)){
    Raw_Net[i,adj[i,]] <- Coefs[[i]][-1]
  }
  if(thresholding == TRUE) {  
    
    # Test for significance
    Sig <- matrix(0, n, n)
    for (i in seq_len(n)){
      Sig[i,adj[i,]] <- p_values[[i]][-1]
    }
    Net <- ifelse(Sig < alpha, Raw_Net, 0 ) 
    
    #And or OR rule:
    if (AND == TRUE) {
      Net <- ifelse(Net != 0 & t(Net != 0), Net, 0)
      Net <- (Net+t(Net))/2 }
    else {
      Net <- (Net+t(Net))/2  
    } }
  else {
    Net <- (Raw_Net + t(Raw_Net)) / 2
  }
  
  # Rescale:
  Trans <- LinTransform(Net, Thresholds, c(0,1), responses)

  # Inverse temperature: parameters are estimated at beta = 1; the Ising
  # measure depends only on beta * graph and beta * thresholds, so dividing
  # by beta yields parameters that, combined with the user's beta, reproduce
  # the data-generating measure:
  return(list(
    graph = Trans$graph / beta,
    thresholds = Trans$thresholds / beta,
    results = Res))
}


# Bivariate DOESNT WORK WHEN 11 COUNT IS LOW!:
EstimateIsingBi <- function(data, responses, beta = 1, ...){
  if (missing(responses)){
    responses <- sort(unique(c(data)))
  }
  
  if (length(responses) != 2){
    stop("Binary data required")
  }
  
  # Rescale data to binary:
  binarize <- function(x, responses){
    x2 <- x
    x2[x==responses[1]] <- 0
    x2[x==responses[2]] <- 1
    x2
  }
  data <- binarize(data, responses)
  
  # Number of variables:
  n <- ncol(data)
  
  # GLM for every pair of nodes:
  Res <- matrix(list(), n, n)
  Net <- matrix(0,n,n)
  ThresholdMat <- matrix(0,n-1,n)
  update <- rep(1, n)
  for (i in seq_len(n)){
    for (j in seq_len(i-1)){
      # Create factor:
      fac <- factor(data[,i] + 2*data[,j])
      Res[[i,j]] <- multinom(fac ~ data[,-c(i,j),drop=FALSE], ...)
      coef <- coef(Res[[i,j]])
      
      if (!is.matrix(coef)){
        coef <- as.matrix(t(coef))
        rownames(coef) <- unique(fac[fac!="0"])
      }
      
      # Add threshold i:
      if (1 %in% fac){
        ThresholdMat[update[i],i] <- coef[rownames(coef)=="1",1]
        update[i] <- update[i] + 1
      }
      
      # update threshold j:
      if (2 %in% fac){
        ThresholdMat[update[j],j] <- coef[rownames(coef)=="2",1]
        update[j] <- update[j]  + 1
      }
      
      # Update network:
      if (3 %in% fac){
        Net[i,j] <- Net[j,i] <- coef[rownames(coef)=="3",1]
      }
      
    }
  }
  
  # Threshold means:
  Thresholds <- colMeans(ThresholdMat)
  
  # Remove thresholds from net:
  Net <- Net - outer(Thresholds,Thresholds,'+')
  diag(Net) <- 0
  
  # Rescale:
  Trans <- LinTransform(Net, Thresholds, c(0,1), responses)

  # Inverse temperature: estimated at beta = 1, rescaled so that
  # beta * graph and beta * thresholds reproduce the data-generating measure:
  return(list(
    graph = Trans$graph / beta,
    thresholds = Trans$thresholds / beta,
    results = Res))
}
# Contrasts matrix:
# Contr <- matrix(c(
#   -1, -1, 1,
#   1, -1, -1,
#   -1, 1, -1,
#   1, 1, 1  
# ),4,3, byrow = TRUE)
# rownames(Contr) <- 0:3
# colnames(Contr) <- 1:3
  
  
  
# As a loglinear model:
EstimateIsingLL <- function(data, responses, beta = 1, adj = matrix(1, ncol(data), ncol(data)), ...){
  if (missing(responses)){
    responses <- sort(unique(c(data)))
  }
  
  if (length(responses) != 2){
    stop("Binary data required")
  }
  
  # Rescale data to binary:
  binarize <- function(x, responses){
    x2 <- x
    x2[x==responses[1]] <- 0
    x2[x==responses[2]] <- 1
    x2
  }
  data <- binarize(data, responses)
  
  # Number of variables:
  n <- ncol(data)
  
  # GLM for every pair of nodes:
  data <- as.data.frame(data)
  for (i in seq_len(ncol(data))){
    data[[i]] <- factor(data[[i]], levels = c(1,0))
  }
  
  # Frequency table:
  Tab <- table(data)
  
  # margins:
  adj <- as.matrix(adj)
  rownames(adj) <- colnames(adj) <- NULL
  Margins <- plyr::alply(which(upper.tri(adj) & adj != 0, arr.ind=TRUE),1,identity)
  # Margins <- alply(t(combn(seq_len(n),2)),1) 
  
  # Estimate
  Res <- loglin(Tab, Margins, param = TRUE, ...)

  # Parameters:
  Params <- Res$param
  Net <- matrix(0,n,n)
  Thresholds <- numeric(n)
  
  Names <- names(data)
  
  # Thresholds:
  for (i in seq_len(n)){
    Thresholds[[i]] <- Params[[Names[[i]]]][[1]]
  }
  
  # Network:
  for (j in seq_len(n)){
    for (i in seq_len(j-1)){
      if (paste0(Names[[i]],".",Names[[j]]) %in% names(Params)){
        Net[i,j]<- Net[j,i] <- Params[[paste0(Names[[i]],".",Names[[j]])]][1,1]        
      }
    }
  }
  
  # Rescale:
  Trans <- LinTransform(Net, Thresholds, c(-1,1), responses)

  # Inverse temperature: estimated at beta = 1, rescaled so that
  # beta * graph and beta * thresholds reproduce the data-generating measure:
  return(list(
    graph = Trans$graph / beta,
    thresholds = Trans$thresholds / beta,
    results = Res))
}

# Tests for the minimum sum-score correction of EstimateIsingUni():
# 'min_sum' refers to sum scores on the ORIGINAL response coding, but the
# row filter used to run after the data were recoded to 0/1, so for data
# coded e.g. c(1,2) it filtered the wrong rows (usually none). The filter
# now runs on the raw data: for node i a row is biased iff the rest score
# is compatible with the truncation when node i takes the high response
# but not when it takes the low response. Checks:
# (a) on 0/1 data the new raw-scale filter retains exactly the rows the
#     old binarized-scale filter (rest != min_sum - 1) retained, per node;
# (b) on c(1,2)-coded data (same data recoded, min_sum shifted accordingly)
#     the estimator drops the corresponding rows and the estimates agree
#     with the 0/1 fit after LinTransform;
# (c) the default min_sum = -Inf drops no rows.
# Plain stopifnot tests, no package dependencies beyond IsingSampler itself.

library("IsingSampler")

## Small seeded 4-node test model ---------------------------------------------
set.seed(42)
N <- 4
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- c(0.4, -0.2, 0.6, 0.2, -0.3, 0.5)
graph <- graph + t(graph)
thresholds <- c(-0.2, 0.1, 0, -0.1)
full <- as.matrix(IsingSampler(1000, graph, thresholds, responses = c(0L, 1L)))

# Mimic sum-score truncated data: keep only rows with sum score >= min_sum:
min_sum <- 2
data01 <- full[rowSums(full) >= min_sum, , drop = FALSE]

# Number of retained rows per fitted node model:
nRows <- function(fit) sapply(fit$results, function(m) length(residuals(m)))

## (a) new raw-scale filter == old filter on 0/1 data --------------------------
# On 0/1 data the binarized data equal the raw data, so the old filter
# (rest != min_sum - 1 on the binarized data) must be reproduced exactly by
# the new condition (rest + 1 >= min_sum & rest + 0 < min_sum on raw data):
uni01 <- EstimateIsingUni(data01, min_sum = min_sum)
for (i in seq_len(N)){
  rest <- rowSums(data01[, -i, drop = FALSE])
  keepOld <- which(rest != min_sum - 1)
  keepNew <- which(!(rest + 1 >= min_sum & rest + 0 < min_sum))
  stopifnot(identical(keepOld, keepNew))
  # The estimator must have fitted node i on exactly these rows:
  stopifnot(nRows(uni01)[i] == length(keepNew))
  # Make the test non-trivial: some rows must actually be dropped:
  stopifnot(length(keepNew) < nrow(data01))
}

## (b) c(1,2)-coded data: same rows dropped, estimates agree -------------------
# Recode the same data to responses c(1,2); each row's sum score shifts by
# the number of nodes, so the equivalent truncation threshold shifts too:
data12 <- data01 + 1
min_sum12 <- min_sum + N
uni12 <- EstimateIsingUni(data12, min_sum = min_sum12)

# Corresponding rows are dropped per node (the old binarized-scale filter
# would have dropped none here, since rest scores of 0/1-binarized data
# never equal min_sum12 - 1 = 5 with only 3 remaining nodes):
stopifnot(identical(nRows(uni12), nRows(uni01)))

# The glm fits run on identical binarized data and identical retained rows,
# so mapping the 0/1 estimates to the (1,2) coding via LinTransform must
# reproduce the (1,2) fit essentially exactly:
trans <- LinTransform(uni01$graph, uni01$thresholds, from = c(0, 1), to = c(1, 2))
stopifnot(isTRUE(all.equal(trans$graph, uni12$graph, tolerance = 1e-6)))
stopifnot(isTRUE(all.equal(trans$thresholds, uni12$thresholds, tolerance = 1e-6)))

## (c) default min_sum = -Inf drops no rows ------------------------------------
uniDef <- EstimateIsingUni(full)
uniInf <- EstimateIsingUni(full, min_sum = -Inf)
stopifnot(all(nRows(uniInf) == nrow(full)))
stopifnot(identical(uniDef$graph, uniInf$graph))
stopifnot(identical(uniDef$thresholds, uniInf$thresholds))

cat("All min_sum tests passed.\n")

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

IsingProcess <- function(nSample, graph, thresholds, beta, responses, delta = NULL) {
    .Call('_IsingSampler_IsingProcess', PACKAGE = 'IsingSampler', nSample, graph, thresholds, beta, responses, delta)
}

IsingSamplerCpp <- function(n, graph, thresholds, beta, nIter, responses, exact, constrain, delta = NULL) {
    .Call('_IsingSampler_IsingSamplerCpp', PACKAGE = 'IsingSampler', n, graph, thresholds, beta, nIter, responses, exact, constrain, delta)
}

H <- function(J, s, h, delta = NULL) {
    .Call('_IsingSampler_H', PACKAGE = 'IsingSampler', J, s, h, delta)
}

f <- function(Y, J, h, delta = NULL) {
    .Call('_IsingSampler_f', PACKAGE = 'IsingSampler', Y, J, h, delta)
}

Hvec <- function(s, Theta, N) {
    .Call('_IsingSampler_Hvec', PACKAGE = 'IsingSampler', s, Theta, N)
}

PseudoLikelihood <- function(x, graph, thresholds, beta, responses, logis) {
    .Call('_IsingSampler_PseudoLikelihood', PACKAGE = 'IsingSampler', x, graph, thresholds, beta, responses, logis)
}

binaryTransform <- function(x, from, to){
  if (missing(from)){
    from <- sort(unique(c(unlist(x))))
  }
  if (length(from) != 2){
    stop("Binary data required")
  }
  
  x2 <- x
  x2[x==from[1]] <- to[1]
  x2[x==from[2]] <- to[2]
  x2
}

LinTransform <- function(graph, thresholds, from = c(0L, 1L), to = c(-1L, 1L), a, b)
{
  stopifnot(!missing(graph) & !missing(thresholds))

  if (length(from) != 2 || length(to) != 2){
    stop("'from' and 'to' must each be vectors of length 2.")
  }

  if (missing(a) != missing(b)){
    stop("Supply both 'a' and 'b', or neither.")
  }

  if (missing(a) & missing(b))
  {
    a <- (to[1]-to[2])/(from[1]-from[2])    
    b <- to[1] - a*from[1]
  }
  
  diag(graph) <- 0
  
  return(list(
    graph = graph/(a^2), 
    thresholds = thresholds/a - (b*rowSums(graph))/a^2))
}

\name{IsingPL}
\alias{IsingPL}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Pseudo-likelihood
}
\description{
Computes the pseudolikelihood of a dataset given an Ising Model.
}
\usage{
IsingPL(x, graph, thresholds, beta, responses = c(0L, 1L))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
A binary dataset
}
  \item{graph}{
Square matrix indicating the weights of the network. Must be symmetrical with 0 as diagonal.
}
  \item{thresholds}{
Vector indicating the thresholds, also known as the external field.
}
  \item{beta}{
Scalar indicating the inverse temperature.
}
  \item{responses}{
Response options. Typically set to \code{c(-1L, 1L)} or \code{c(0L, 1L)} (default). Must be integers!
}
}

\value{
The pseudolikelihood
}
\author{
Sacha Epskamp (mail@sachaepskamp.com)
}

\name{IsingSampler}
\alias{IsingSampler}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Sample states from the Ising model
}
\description{
This function samples states from the Ising model using one of three methods. See details.
}
\usage{
IsingSampler(n, graph, thresholds, beta = 1, nIter = 100, responses = c(0L, 1L),
    method = c("MH", "CFTP", "direct"), CFTPretry = 10, constrain, delta = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{n}{
Number of states to draw
}
  \item{graph}{
Square matrix indicating the weights of the network. Must be symmetrical with 0 as diagonal.
}
  \item{thresholds}{
Vector indicating the thresholds, also known as the external field.
}
  \item{beta}{
Scalar indicating the inverse temperature.
}
  \item{nIter}{
Number of iterations in the Metropolis and exact sampling methods.
}
  \item{responses}{
Response options. Typically set to \code{c(-1L, 1L)} or \code{c(0L, 1L)} (default). Two or more response options are supported, and they may be any numeric values, including non-integers (e.g. \code{c(-1L, 0L, 1L)}, \code{1:5} or \code{seq(-1, 1, by = 0.5)}). With more than two options the conditional distribution of each node is the categorical (softmax) generalization of the binary Ising conditional. The pairwise interaction always enters the Hamiltonian as \code{J_ij * s_i * s_j}, i.e. through the (numeric) product of the response values, and each node has a single threshold. When all response options are integer-valued the result is an integer matrix (as before); otherwise it is a numeric (double) matrix.
}
  \item{method}{
The sampling method to use. Must be \code{"MH"}, \code{"CFTP"} or \code{"direct"}. See details.
}
 \item{CFTPretry}{
 The amount of times a sample from CFTP may be retried. If after 100 couplings from the past the chain still results in \code{NA} values the chain is reset with different random numbers. Be aware that data that requies a lot of CFTP resets might not resemble exact samples anymore.
 }
 \item{constrain}{A (number of samples) by (number of nodes) matrix with samples that need be constrained; \code{NA} indicates that the sample is unconstrained. Defaults to a matrix of \code{NA}s. Non-\code{NA} values must be members of \code{responses}. Constraints are honored by all three sampling methods: constrained (clamped) nodes are fixed to the given value and the remaining nodes are sampled from the conditional distribution given the clamped nodes (for \code{"MH"} and \code{"CFTP"} the clamped nodes are simply never updated; for \code{"direct"} the state space is restricted to matching states and the probabilities are renormalized).}
  \item{delta}{
Optional per-node quadratic (Blume-Capel single-ion / crystal-field) term added to the Hamiltonian as \code{delta_i * s_i^2}. Either a single value (recycled over nodes) or a vector with one value per node. The default \code{0} reproduces the ordinary Ising model. A positive \code{delta} favours response options near zero (the middle category); a negative \code{delta} favours the extreme categories. This term only has an identifiable effect with more than two response options; with two response options it is absorbed into the thresholds (and is constant, hence has no effect, for symmetric responses such as \code{c(-1, 1)}). See \code{\link{BlumeCapelSampler}}.
}
}
\details{
This function uses one of three sampling methods. \code{"MH"} can be used to sample using a Metropolis-Hastings algorithm. The chain is initiated with random values from the response options, then for each iteration each node is redrawn from its full-conditional distribution given all other nodes and parameters. For two response options this is the usual binary update; for more than two response options the node is drawn from the categorical (softmax) full-conditional over all response options. Typically, 100 of such iterations should suffice for the chain to converge, though strongly coupled models with more than two response options may require a larger \code{nIter}.

The second method, \code{"CFTP"} enhances the Metropolis-Hastings algorithm with Coupling from the Past (CFTP; Murray, 2007) to draw exact samples from the distribution. This is slower than the default Metropolis-Hastings but guarantees exact samples. However, it does depend on the graph structure and the number of nodes if these exact samples can be obtained in feasable time. \code{"CFTP"} is currently only implemented for two response options; for more than two response options use \code{"MH"} or \code{"direct"}.

The third option, \code{"direct"}, simply computes for every possibly state the probability and draws samples directly from the distribution of states by using these probabilities. This also guarantees exact samples, but quickly becomes intractible (roughly above 10 nodes).
}
\value{
A matrix containing samples of states.
}
\references{
Murray, I. (2007). Advances in Markov chain Monte Carlo methods.}
\author{
Sacha Epskamp (mail@sachaepskamp.com)
}

\seealso{
\code{\link{IsingSampler-package}} for examples; \code{\link{BlumeCapelSampler}} for the Blume-Capel (quadratic / crystal-field) extension.
}
\examples{
## See IsingSampler-package help page for the main usage examples.

## Constraining (clamping) nodes via the 'constrain' argument -----------------
## 'constrain' is a (number of samples) by (number of nodes) matrix; NA marks a
## free (unconstrained) node, a non-NA value clamps that node to that response.
N <- 3
graph <- matrix(0.5, N, N)
diag(graph) <- 0
thresholds <- rep(0, N)

# Clamp node 1 to 1 in every one of the 10 samples, leave nodes 2 and 3 free:
constrain <- matrix(NA_integer_, 10, N)
constrain[, 1] <- 1L

# All three methods honor the constraint (node 1 is always 1):
set.seed(1)
sMH     <- IsingSampler(10, graph, thresholds, method = "MH",     constrain = constrain)
sCFTP   <- IsingSampler(10, graph, thresholds, method = "CFTP",   constrain = constrain)
sDirect <- IsingSampler(10, graph, thresholds, method = "direct", constrain = constrain)

all(sMH[, 1] == 1L)      # TRUE
all(sCFTP[, 1] == 1L)    # TRUE
all(sDirect[, 1] == 1L)  # TRUE
}

\name{NodeInformation}
\alias{NodeInformation}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Mutual information between each node and the rest of the network
}
\description{
Computes, for each node of a (tractable) Ising model, the mutual information
\eqn{I(\text{node}; \text{rest of network})} between that node and all remaining
nodes. This quantifies how much knowing the state of a single node reduces the
uncertainty (Shannon entropy) about the joint state of the remaining nodes.
}
\usage{
NodeInformation(graph, thresholds, beta = 1, base = 2,
          responses = c(0L, 1L), delta = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{graph}{
Weights matrix
}
  \item{thresholds}{
Thresholds vector
}
  \item{beta}{
Inverse temperature
}
  \item{base}{
Base of the logarithm
}
  \item{responses}{
Vector of outcome responses.
}
  \item{delta}{
Optional per-node quadratic (Blume-Capel) term added to the Hamiltonian as \code{delta_i * s_i^2}; a single value (recycled over nodes) or one value per node. The default \code{0} corresponds to the ordinary Ising model. See \code{\link{BlumeCapelSampler}}.
}
}
\details{
For each node \eqn{i} the returned value is computed via \code{\link{IsingEntropy}}
as the entropy of the rest of the network with node \eqn{i} marginalised out,
minus the entropy of the rest of the network conditional on node \eqn{i}:
\deqn{I_i = H(\text{rest}) - H(\text{rest} \mid \text{node}_i).}
This is the standard decomposition of the mutual information
\eqn{I(\text{node}_i; \text{rest})}. As mutual information it is non-negative
(up to numerical error). The computation is only tractable for small networks
(roughly up to 10 nodes), since it enumerates the full state space via
\code{\link{IsingLikelihood}}.
}
\value{
A numeric vector of length \code{ncol(graph)}, giving the mutual information (in
units set by \code{base}) between each node and the remaining nodes of the
network.
}
\author{
Sacha Epskamp <mail@sachaepskamp.com>
}

\seealso{
\code{\link{IsingEntropy}}
}
\examples{
# Small 3-node network:
N <- 3
graph <- matrix(0, N, N)
graph[upper.tri(graph)] <- c(0.5, 0.2, 0.8)
graph <- graph + t(graph)
thresholds <- rep(0, N)

# Mutual information of each node with the rest of the network:
NodeInformation(graph, thresholds)
}

\name{IsingStateProb}
\alias{IsingStateProb}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Likelihood of single state from tractible Ising model.
}
\description{
This function returns the likelihood of a single possible state. Is only tractible up to rougly 10 nodes.
}
\usage{
IsingStateProb(s, graph, thresholds, beta, responses = c(0L, 1L), delta = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{s}{
Vector contaning the state to evaluate.
}
  \item{graph}{
Square matrix indicating the weights of the network. Must be symmetrical with 0 as diagonal.
}
  \item{thresholds}{
Vector indicating the thresholds, also known as the external field.
}
  \item{beta}{
Scalar indicating the inverse temperature.
}
  \item{responses}{
Response options. Typically set to \code{c(-1L, 1L)} or \code{c(0L, 1L)} (default). Two or more numeric response options are supported, including non-integer values such as \code{seq(-1, 1, by = 0.5)}.
}
  \item{delta}{
Optional per-node quadratic (Blume-Capel) term added to the Hamiltonian as \code{delta_i * s_i^2}; a single value (recycled over nodes) or one value per node. The default \code{0} corresponds to the ordinary Ising model. See \code{\link{BlumeCapelSampler}}.
}
}

\author{
Sacha Epskamp (mail@sachaepskamp.com)
}

\name{BlumeCapelSampler}
\alias{BlumeCapelSampler}
\title{
Sample states from the Blume-Capel model
}
\description{
Samples states from the Blume-Capel model, an Ising model with an additional on-site quadratic ("single-ion anisotropy" or "crystal-field") term. This is a thin wrapper around \code{\link{IsingSampler}} that sets ordered (spin-1) response options by default and requires the quadratic parameter \code{delta}.
}
\usage{
BlumeCapelSampler(n, graph, thresholds, delta, beta = 1, nIter = 100,
    responses = c(-1L, 0L, 1L), method = c("MH", "direct"),
    CFTPretry = 10, constrain)
}
\arguments{
  \item{n}{
Number of states to draw.
}
  \item{graph}{
Square matrix indicating the weights of the network. Must be symmetrical with 0 as diagonal.
}
  \item{thresholds}{
Vector indicating the thresholds (external field). A single value is recycled over all nodes.
}
  \item{delta}{
The quadratic (Blume-Capel) parameter, entering the Hamiltonian as \code{delta_i * s_i^2}. Either a single value (recycled over all nodes) or one value per node. Must be supplied. A positive \code{delta} favours the middle category; a negative \code{delta} favours the extreme categories.
}
  \item{beta}{
Scalar indicating the inverse temperature.
}
  \item{nIter}{
Number of iterations in the Metropolis sampler.
}
  \item{responses}{
Ordered response options, defaulting to the spin-1 values \code{c(-1L, 0L, 1L)}. Other (numeric) response options may be supplied, but the \code{delta} term only has an identifiable effect with more than two options.
}
  \item{method}{
The sampling method, either \code{"MH"} (Metropolis-Hastings, default) or \code{"direct"}. Exact sampling (\code{"CFTP"}) is implemented for two response options only and is therefore not available here; \code{delta} has no identifiable effect on binary responses in any case.
}
  \item{CFTPretry}{
Passed to \code{\link{IsingSampler}}; unused for the methods offered here.
}
  \item{constrain}{
A (number of samples) by (number of nodes) matrix with samples that need be constrained; \code{NA} indicates that the sample is unconstrained.
}
}
\details{
The Blume-Capel Hamiltonian implemented here is
\deqn{H(s) = -\sum_i \tau_i s_i - \sum_{i<j} \omega_{ij} s_i s_j + \sum_i \delta_i s_i^2,}
with thresholds \eqn{\tau}, network weights \eqn{\omega} (\code{graph}) and quadratic terms \eqn{\delta} (\code{delta}); states are drawn with probability proportional to \eqn{\exp(-\beta H(s))}. Setting \code{delta = 0} recovers the ordinary (multi-level) Ising model, so \code{BlumeCapelSampler(..., delta = 0)} is equivalent to \code{\link{IsingSampler}}.
}
\value{
A matrix containing samples of states.
}
\author{
Sacha Epskamp (mail@sachaepskamp.com)
}
\seealso{
\code{\link{IsingSampler}}, \code{\link{IsingLikelihood}}
}
\examples{
\dontrun{
# A small ferromagnetic network:
P <- 5
W <- matrix(0.5, P, P); diag(W) <- 0

# Low crystal field: ordered (mostly +1 / -1):
X0 <- BlumeCapelSampler(1000, W, thresholds = 0, delta = 0, beta = 1)

# High crystal field: disordered "0" phase (mostly the middle category):
X3 <- BlumeCapelSampler(1000, W, thresholds = 0, delta = 3, beta = 1)

mean(X0 == 0)  # low
mean(X3 == 0)  # high
}
}

\name{LinTransform}
\alias{LinTransform}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Transform parameters for linear transformations on response catagories
}
\description{
This function is mainly used to translate parameters estimated with response options set to 0 and 1 to a model in which the response options are -1 and 1, but can be used for any linear transformation of response options.
}
\usage{
LinTransform(graph, thresholds, from = c(0L, 1L), to = c(-1L, 1L), a, b)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{graph}{
A matrix containing an Ising graph
}
  \item{thresholds}{
A vector containing thresholds
}
  \item{from}{
The original response encoding
}
  \item{to}{
The response encoding to transform to
}
  \item{a}{
The slope of the transformation. Overwrites \code{to}.
}
  \item{b}{
The intercept of the transformation. Overwrites \code{to}.
}
}

\author{
Sacha Epskamp <sacha.epskamp@gmail.com>
}

\examples{
N <- 4 # Number of nodes

# Ising parameters:
Graph <- matrix(sample(0:1,N^2,TRUE,prob = c(0.7, 0.3)),N,N) * rnorm(N^2)
Graph <- pmax(Graph,t(Graph)) / N
diag(Graph) <- 0
Thresh <- -(rnorm(N)^2)
Beta <- 1

p1 <- IsingLikelihood(Graph, Thresh, Beta, c(0,1))

a <- 2
b <- -1

# p2 <- IsingLikelihood(Graph/(a^2), Thresh/a - (b*rowSums(Graph))/a^2, Beta, c(-1,1))

p2 <- IsingLikelihood(LinTransform(Graph,Thresh)$graph, 
                      LinTransform(Graph,Thresh)$thresholds , 
                      Beta, c(-1,1))
LinTransform

round(cbind(p1[,1],p2[,1]),5)

plot(p1[,1],p2[,1])
abline(0,1)
}

\name{IsingEntropy}
\alias{IsingEntropy}
\alias{IsingEntrophy}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Entropy of the Ising Model
}
\description{
Returns (marginal/conditional) Shannon entropy of the Ising model.
}
\usage{
IsingEntropy(graph, thresholds, beta = 1, conditional = numeric(0),
          marginalize = numeric(0), base = 2, responses = c(0L, 1L),
          delta = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{graph}{
Weights matrix
}
  \item{thresholds}{
Thresholds vector
}
  \item{beta}{
Inverse temperature
}
  \item{conditional}{
Indices of nodes to condition on
}
  \item{marginalize}{
Indices of nodes to marginalize over
}
  \item{base}{
Base of the logarithm
}
  \item{responses}{
Vector of outcome responses.
}
  \item{delta}{
Optional per-node quadratic (Blume-Capel) term added to the Hamiltonian as \code{delta_i * s_i^2}; a single value (recycled over nodes) or one value per node. The default \code{0} corresponds to the ordinary Ising model. See \code{\link{BlumeCapelSampler}}.
}
}
\note{
The old (misspelled) name \code{IsingEntrophy} is deprecated; use
\code{IsingEntropy} instead. \code{IsingEntrophy} is retained as a thin
wrapper that emits a deprecation warning and forwards to \code{IsingEntropy}.
}

\author{
Sacha Epskamp <mail@sachaepskamp.com>
}

\name{EstimateIsing}
\alias{EstimateIsing}
\alias{EstimateIsingPL}
\alias{EstimateIsingUni}
\alias{EstimateIsingBi}
\alias{EstimateIsingLL}

%- Also NEED an '\alias' for EACH other topic documented here.
\title{
non-regularized estimation methods for the Ising Model
}
\description{
This function can be used for several non-regularized estimation methods of the Ising Model. See details.
}
\usage{
EstimateIsing(data, responses, beta = 1, method = c("uni", "pl",
                 "bi", "ll"), adj = matrix(1, ncol(data), ncol(data)),
                 ...)
EstimateIsingUni(data, responses, beta = 1, adj = matrix(1, ncol(data),
                 ncol(data)), min_sum = -Inf, thresholding = FALSE, alpha = 0.01, 
                 AND = TRUE, ...)
EstimateIsingBi(data, responses, beta = 1, ...)
EstimateIsingPL(data, responses, beta = 1, ...)
EstimateIsingLL(data, responses, beta = 1, adj = matrix(1, ncol(data),
                 ncol(data)), ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{data}{
 Data frame with binary responses to estimate the Ising model over
}
  \item{responses}{
 Vector of length two indicating the response coding (usually \code{c(0L, 1L)} or \code{c(-1L, 1L)}). When supplied, it must match the coding actually present in the data; if either response option is absent from the data a warning is given (and the results are likely meaningless). If missing, the coding is taken from the data as \code{sort(unique(c(data)))}.
}
  \item{beta}{
 Inverse temperature parameter. The Ising measure depends on the parameters only through the products \code{beta * graph} and \code{beta * thresholds}, so the parameters are only identified up to this scaling. All estimators return \code{graph} and \code{thresholds} such that, combined with the supplied \code{beta}, they reproduce the data-generating distribution; equivalently, the returned parameters equal the estimates obtained with \code{beta = 1} divided by \code{beta}. For \code{pl} this is achieved by maximizing the pseudolikelihood at the supplied \code{beta}; for \code{uni}, \code{bi} and \code{ll} the parameters are estimated at \code{beta = 1} and divided by \code{beta}. The raw \code{results} element is not rescaled.
}
  \item{method}{
  The method to be used. \code{pl} uses pseudolikelihood estimation, \code{uni} sequential univariate regressions, \code{bi} bivariate regressions and \code{ll} estimates the Ising model as a loglinear model.
}
\item{adj}{
Adjacency matrix of the Ising model.
}
  \item{min_sum}{ The minimum sum score that is artifically possible in the dataset. Defaults to -Inf. Set this only if you know a lower sum score is not possible in the data, for example due to selection bias.}
    \item{thresholding}{Logical, should the model be thresholded for significance?}
    \item{alpha}{Alpha level used in thresholding}
    \item{AND}{Logical, should an AND-rule (both regressions need to be significant) or OR-rule (one of the regressions needs to be significant) be used?}
    \item{\dots}{
 Arguments sent to estimator functions
}
}
\details{
The following algorithms can be used (see Epskamp, Maris, Waldorp, Borsboom; in press).
\describe{
\item{\code{pl}}{Estimates the Ising model by maximizing the pseudolikelihood (Besag, 1975).}
\item{\code{uni}}{Estimates the Ising model by computing univariate logistic regressions of each node on all other nodes. This leads to a single estimate for each threshold and two estimates for each network parameter. The two estimates are averaged to produce the final network. Uses \code{\link{glm}}.}
\item{\code{bi}}{Estimates the Ising model using multinomial logistic regression of each pair of nodes on all other nodes. This leads to a single estimate of each network parameter and $p$ estimates of each threshold parameter. Uses \code{\link[nnet]{multinom}}.}
\item{\code{ll}}{Estimates the Ising model by phrasing it as a loglinear model with at most pairwise interactions. Uses \code{\link{loglin}}.}
}
}
\value{
A list containing the estimation results:
\item{graph}{The estimated network, scaled such that \code{beta * graph} corresponds to the data-generating measure}
\item{thresholds}{The estimated thresholds, scaled such that \code{beta * thresholds} corresponds to the data-generating measure}
\item{results}{The results object used in the analysis (on the raw estimation scale, not rescaled by \code{beta})}
%% ...
}
\references{
Epskamp, S., Maris, G., Waldorp, L. J., and Borsboom, D. (in press). Network Psychometrics. To appear in: Irwing, P., Hughes, D., and Booth, T. (Eds.), Handbook of Psychometrics. New York: Wiley.

Besag, J. (1975), Statistical analysis of non-lattice data. The statistician, 24, 179-195.
}
\note{
Method \code{"bi"} (\code{EstimateIsingBi}), which fits a multinomial logistic
regression for each pair of nodes, can be unreliable when the joint
high--high response category (both nodes taking the high response, i.e. the
\code{(1, 1)} pair) is rarely or never observed. In that situation the
corresponding multinomial cell is (near-)empty and the estimates for the
affected parameters become unstable or fail to be identified. Prefer another
method (for example \code{"pl"} or \code{"uni"}) when such response pairs are
sparse.
}
\author{
Sacha Epskamp (mail@sachaepskamp.com)
}
\examples{
# Input:
N <- 5 # Number of nodes
nSample <- 500 # Number of samples

# Ising parameters:
Graph <- matrix(sample(0:1,N^2,TRUE,prob = c(0.7, 0.3)),N,N) * rnorm(N^2)
Graph <- Graph + t(Graph)
diag(Graph) <- 0
Thresholds <- rep(0,N)
Beta <- 1

# Response options (0,1 or -1,1):
Resp <- c(0L,1L)
Data <- IsingSampler(nSample,Graph, Thresholds)

# Pseudolikelihood:
resPL <- EstimateIsing(Data, method = "pl")
cor(Graph[upper.tri(Graph)], resPL$graph[upper.tri(resPL$graph)])

# Univariate logistic regressions:
resUni <- EstimateIsing(Data, method = "uni")
cor(Graph[upper.tri(Graph)], resUni$graph[upper.tri(resUni$graph)])

# bivariate logistic regressions:
resBi <- EstimateIsing(Data, method = "bi")
cor(Graph[upper.tri(Graph)], resBi$graph[upper.tri(resBi$graph)])

# Loglinear model:
resLL <- EstimateIsing(Data, method = "ll")
cor(Graph[upper.tri(Graph)], resLL$graph[upper.tri(resLL$graph)])
}
